package iotcloud;
import java.util.HashMap;
import java.util.Map;
import java.util.Iterator;
import java.util.HashSet;
import java.util.Arrays;
import java.util.Vector;
import java.util.Random;
import java.util.Queue;
import java.util.LinkedList;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;
import java.util.Collection;
import java.util.Collections;


/**
 * IoTTable data structure.  Provides client inferface.
 * @author Brian Demsky
 * @version 1.0
 */

final public class Table {
	private int numslots;	//number of slots stored in buffer

	//table of key-value pairs
	//private HashMap<IoTString, KeyValue> table = new HashMap<IoTString, KeyValue>();

	// machine id -> (sequence number, Slot or LastMessage); records last message by each client
	private HashMap<Long, Pair<Long, Liveness> > lastmessagetable = new HashMap<Long, Pair<Long, Liveness> >();
	// machine id -> ...
	private HashMap<Long, HashSet<RejectedMessage> > watchlist = new HashMap<Long, HashSet<RejectedMessage> >();
	private Vector<Long> rejectedmessagelist = new Vector<Long>();
	private SlotBuffer buffer;
	private CloudComm cloud;
	private long sequencenumber; //Largest sequence number a client has received
	private long localmachineid;
	private TableStatus lastTableStatus;
	static final int FREE_SLOTS = 10; //number of slots that should be kept free
	static final int SKIP_THRESHOLD = 10;
	private long liveslotcount = 0;
	private int chance;
	static final double RESIZE_MULTIPLE = 1.2;
	static final double RESIZE_THRESHOLD = 0.75;
	static final int REJECTED_THRESHOLD = 5;
	private int resizethreshold;
	private long lastliveslotseqn;	//smallest sequence number with a live entry
	private Random random = new Random();
	private long lastUncommittedTransaction = 0;

	private int smallestTableStatusSeen = -1;
	private int largestTableStatusSeen = -1;

	private PendingTransaction pendingTransBuild = null; // Pending Transaction used in building
	private Queue<PendingTransaction> pendingTransQueue = null; // Queue of pending transactions
	private Map<Long, Commit> commitMap = null; // List of all the most recent live commits
	private Map<Long, Abort> abortMap = null; // Set of the live aborts
	private Map<IoTString, Commit> committedMapByKey = null; // Table of committed KV
	private Map<IoTString, KeyValue> commitedTable = null; // Table of committed KV
	private Map<IoTString, KeyValue> speculativeTable = null; // Table of speculative KV
	private Map<Long, Transaction> uncommittedTransactionsMap = null;
	private Map<IoTString, Long> arbitratorTable = null; // Table of arbitrators
	private Map<IoTString, NewKey> newKeyTable = null; // Table of speculative KV
	private Map<Long, Commit> newCommitMap = null; // Map of all the new commits
	private Map<Long, Long> lastCommitSeenSeqNumMap = null; // sequence number of the last commit that was seen grouped by arbitrator
	private Map<Long, Long> lastAbortSeenSeqNumMap = null; // sequence number of the last commit that was seen grouped by arbitrator



	public Table(String baseurl, String password, long _localmachineid) {
		localmachineid = _localmachineid;
		buffer = new SlotBuffer();
		numslots = buffer.capacity();
		setResizeThreshold();
		sequencenumber = 0;
		cloud = new CloudComm(this, baseurl, password);
		lastliveslotseqn = 1;

		setupDataStructs();
	}

	public Table(CloudComm _cloud, long _localmachineid) {
		localmachineid = _localmachineid;
		buffer = new SlotBuffer();
		numslots = buffer.capacity();
		setResizeThreshold();
		sequencenumber = 0;
		cloud = _cloud;

		setupDataStructs();
	}

	private void setupDataStructs() {
		pendingTransQueue = new LinkedList<PendingTransaction>();
		commitMap = new HashMap<Long, Commit>();
		abortMap = new HashMap<Long, Abort>();
		committedMapByKey = new HashMap<IoTString, Commit>();
		commitedTable = new HashMap<IoTString, KeyValue>();
		speculativeTable = new HashMap<IoTString, KeyValue>();
		uncommittedTransactionsMap = new HashMap<Long, Transaction>();
		arbitratorTable = new HashMap<IoTString, Long>();
		newKeyTable = new HashMap<IoTString, NewKey>();
		newCommitMap = new HashMap<Long, Commit>();
		lastCommitSeenSeqNumMap = new HashMap<Long, Long>();
		lastAbortSeenSeqNumMap = new HashMap<Long, Long>();
	}

	public void rebuild() throws ServerException {
		Slot[] newslots = cloud.getSlots(sequencenumber + 1);
		validateandupdate(newslots, true);
	}

	// // TODO: delete method
	// public void printSlots() {
	// 	long o = buffer.getOldestSeqNum();
	// 	long n = buffer.getNewestSeqNum();

	// 	int[] types = new int[10];

	// 	int num = 0;

	// 	int livec = 0;
	// 	int deadc = 0;
	// 	for (long i = o; i < (n + 1); i++) {
	// 		Slot s = buffer.getSlot(i);

	// 		Vector<Entry> entries = s.getEntries();

	// 		for (Entry e : entries) {
	// 			if (e.isLive()) {
	// 				int type = e.getType();
	// 				types[type] = types[type] + 1;
	// 				num++;
	// 				livec++;
	// 			} else {
	// 				deadc++;
	// 			}
	// 		}
	// 	}

	// 	for (int i = 0; i < 10; i++) {
	// 		System.out.println(i + "    " + types[i]);
	// 	}
	// 	System.out.println("Live count:   " + livec);
	// 	System.out.println("Dead count:   " + deadc);
	// 	System.out.println("Old:   " + o);
	// 	System.out.println("New:   " + n);
	// 	System.out.println("Size:   " + buffer.size());
	// 	System.out.println("Commits Map:   " + commitedTable.size());
	// 	System.out.println("Commits List:   " + commitMap.size());
	// }

	public IoTString getCommitted(IoTString key) {
		KeyValue kv = commitedTable.get(key);
		if (kv != null) {
			return kv.getValue();
		} else {
			return null;
		}
	}

	public IoTString getSpeculative(IoTString key) {
		KeyValue kv = speculativeTable.get(key);
		if (kv != null) {
			return kv.getValue();
		} else {
			return null;
		}
	}

	public IoTString getCommittedAtomic(IoTString key) {
		KeyValue kv = commitedTable.get(key);

		if (arbitratorTable.get(key) == null) {
			throw new Error("Key not Found.");
		}

		// Make sure new key value pair matches the current arbitrator
		if (!pendingTransBuild.checkArbitrator(arbitratorTable.get(key))) {
			// TODO: Maybe not throw en error
			throw new Error("Not all Key Values Match Arbitrator.");
		}

		if (kv != null) {
			pendingTransBuild.addKVGuard(new KeyValue(key, kv.getValue()));
			return kv.getValue();
		} else {
			pendingTransBuild.addKVGuard(new KeyValue(key, null));
			return null;
		}
	}

	public IoTString getSpeculativeAtomic(IoTString key) {

		if (arbitratorTable.get(key) == null) {
			throw new Error("Key not Found.");
		}

		// Make sure new key value pair matches the current arbitrator
		if (!pendingTransBuild.checkArbitrator(arbitratorTable.get(key))) {
			// TODO: Maybe not throw en error
			throw new Error("Not all Key Values Match Arbitrator.");
		}

		KeyValue kv = speculativeTable.get(key);
		if (kv == null) {
			kv = commitedTable.get(key);
		}

		if (kv != null) {
			pendingTransBuild.addKVGuard(new KeyValue(key, kv.getValue()));
			return kv.getValue();
		} else {
			pendingTransBuild.addKVGuard(new KeyValue(key, null));
			return null;
		}
	}

	public Long getArbitrator(IoTString key) {
		return arbitratorTable.get(key);
	}

	public void initTable() throws ServerException {
		cloud.setSalt();//Set the salt
		Slot s = new Slot(this, 1, localmachineid);
		TableStatus status = new TableStatus(s, numslots);
		s.addEntry(status);
		Slot[] array = cloud.putSlot(s, numslots);
		if (array == null) {
			array = new Slot[] {s};
			/* update data structure */
			validateandupdate(array, true);
		} else {
			throw new Error("Error on initialization");
		}
	}

	public String toString() {
		String retString = " Committed Table: \n";
		retString += "---------------------------\n";
		retString += commitedTable.toString();

		retString += "\n\n";

		retString += " Speculative Table: \n";
		retString += "---------------------------\n";
		retString += speculativeTable.toString();

		return retString;
	}

	public void startTransaction() {
		// Create a new transaction, invalidates any old pending transactions.
		pendingTransBuild = new PendingTransaction();
	}

	public void commitTransaction() throws ServerException {

		if (pendingTransBuild.getKVUpdates().size() == 0) {
			// If no updates are made then there is no point inserting into the chain
			return;
		}

		// Add the pending transaction to the queue
		pendingTransQueue.add(pendingTransBuild);

		// Delete since already inserted
		pendingTransBuild = new PendingTransaction();

		while (!pendingTransQueue.isEmpty()) {
			if (tryput( pendingTransQueue.peek(), false)) {
				pendingTransQueue.poll();
			}
		}
	}

	public void addKV(IoTString key, IoTString value) {

		if (arbitratorTable.get(key) == null) {
			throw new Error("Key not Found.");
		}

		// Make sure new key value pair matches the current arbitrator
		if (!pendingTransBuild.checkArbitrator(arbitratorTable.get(key))) {
			// TODO: Maybe not throw en error
			throw new Error("Not all Key Values Match Arbitrator.");
		}

		KeyValue kv = new KeyValue(key, value);
		pendingTransBuild.addKV(kv);
	}

	public void update()  throws ServerException {

		Slot[] newslots = cloud.getSlots(sequencenumber + 1);

		validateandupdate(newslots, false);


		if (!pendingTransQueue.isEmpty()) {
			// We have a pending transaction so do full insertion

			while (!pendingTransQueue.isEmpty()) {
				if (tryput( pendingTransQueue.peek(), false)) {
					pendingTransQueue.poll();
				}
			}
		} else {
			// We dont have a pending transaction so do minimal effort
			if (uncommittedTransactionsMap.keySet().size() > 0) {

				boolean doEnd = false;
				boolean needResize = false;
				while (!doEnd && (uncommittedTransactionsMap.keySet().size() > 0)) {
					boolean resize = needResize;
					needResize = false;

					Slot s = new Slot(this, sequencenumber + 1, localmachineid, buffer.getSlot(sequencenumber).getHMAC());
					int newsize = 0;
					if (liveslotcount > resizethreshold) {
						resize = true; //Resize is forced
					}

					if (resize) {
						newsize = (int) (numslots * RESIZE_MULTIPLE);
						TableStatus status = new TableStatus(s, newsize);
						s.addEntry(status);
					}

					doRejectedMessages(s);

					ThreeTuple<Boolean, Boolean, Long> retTup = doMandatoryResuce(s, resize);

					// Resize was needed so redo call
					if (retTup.getFirst()) {
						needResize = true;
						continue;
					}

					// Extract working variables
					boolean seenliveslot = retTup.getSecond();
					long seqn = retTup.getThird();

					// Did need to arbitrate
					doEnd = !doArbitration(s);

					doOptionalRescue(s, seenliveslot, seqn, resize);

					int max = 0;
					if (resize) {
						max = newsize;
					}

					Slot[] array = cloud.putSlot(s, max);
					if (array == null) {
						array = new Slot[] {s};
						rejectedmessagelist.clear();
					}	else {
						if (array.length == 0)
							throw new Error("Server Error: Did not send any slots");
						rejectedmessagelist.add(s.getSequenceNumber());
						doEnd = false;
					}

					/* update data structure */
					validateandupdate(array, true);
				}
			}
		}
	}

	public boolean createNewKey(IoTString keyName, long machineId)  throws ServerException {

		while (true) {
			if (arbitratorTable.get(keyName) != null) {
				// There is already an arbitrator
				return false;
			}

			if (tryput(keyName, machineId, false)) {
				// If successfully inserted
				return true;
			}
		}
	}

	public void decrementLiveCount() {
		liveslotcount--;
	}

	private void setResizeThreshold() {
		int resize_lower = (int) (RESIZE_THRESHOLD * numslots);
		resizethreshold = resize_lower - 1 + random.nextInt(numslots - resize_lower);
	}

	private boolean tryput(PendingTransaction pendingTrans, boolean resize)  throws ServerException {
		Slot s = new Slot(this, sequencenumber + 1, localmachineid, buffer.getSlot(sequencenumber).getHMAC());

		int newsize = 0;
		if (liveslotcount > resizethreshold) {
			resize = true; //Resize is forced
		}

		if (resize) {
			newsize = (int) (numslots * RESIZE_MULTIPLE);
			TableStatus status = new TableStatus(s, newsize);
			s.addEntry(status);
		}

		doRejectedMessages(s);

		ThreeTuple<Boolean, Boolean, Long> retTup = doMandatoryResuce(s, resize);

		// Resize was needed so redo call
		if (retTup.getFirst()) {
			return tryput(pendingTrans, true);
		}

		// Extract working variables
		boolean seenliveslot = retTup.getSecond();
		long seqn = retTup.getThird();

		doArbitration(s);

		Transaction trans = new Transaction(s,
		                                    s.getSequenceNumber(),
		                                    localmachineid,
		                                    pendingTrans.getArbitrator(),
		                                    pendingTrans.getKVUpdates(),
		                                    pendingTrans.getKVGuard());
		boolean insertedTrans = false;
		if (s.hasSpace(trans)) {
			s.addEntry(trans);
			insertedTrans = true;
		}

		doOptionalRescue(s, seenliveslot, seqn, resize);
		return doSendSlotsAndInsert(s, insertedTrans, resize, newsize);
	}

	private boolean tryput(IoTString keyName, long arbMachineid, boolean resize)  throws ServerException {
		Slot s = new Slot(this, sequencenumber + 1, localmachineid, buffer.getSlot(sequencenumber).getHMAC());
		int newsize = 0;
		if (liveslotcount > resizethreshold) {
			resize = true; //Resize is forced
		}

		if (resize) {
			newsize = (int) (numslots * RESIZE_MULTIPLE);
			TableStatus status = new TableStatus(s, newsize);
			s.addEntry(status);
		}

		doRejectedMessages(s);
		ThreeTuple<Boolean, Boolean, Long> retTup = doMandatoryResuce(s, resize);

		// Resize was needed so redo call
		if (retTup.getFirst()) {
			return tryput(keyName, arbMachineid, true);
		}

		// Extract working variables
		boolean seenliveslot = retTup.getSecond();
		long seqn = retTup.getThird();

		doArbitration(s);

		NewKey newKey = new NewKey(s, keyName, arbMachineid);

		boolean insertedNewKey = false;
		if (s.hasSpace(newKey)) {
			s.addEntry(newKey);
			insertedNewKey = true;
		}

		doOptionalRescue(s, seenliveslot, seqn, resize);
		return doSendSlotsAndInsert(s, insertedNewKey, resize, newsize);
	}

	private void doRejectedMessages(Slot s) {
		if (! rejectedmessagelist.isEmpty()) {
			/* TODO: We should avoid generating a rejected message entry if
			 * there is already a sufficient entry in the queue (e.g.,
			 * equalsto value of true and same sequence number).  */

			long old_seqn = rejectedmessagelist.firstElement();
			if (rejectedmessagelist.size() > REJECTED_THRESHOLD) {
				long new_seqn = rejectedmessagelist.lastElement();
				RejectedMessage rm = new RejectedMessage(s, localmachineid, old_seqn, new_seqn, false);
				s.addEntry(rm);
			} else {
				long prev_seqn = -1;
				int i = 0;
				/* Go through list of missing messages */
				for (; i < rejectedmessagelist.size(); i++) {
					long curr_seqn = rejectedmessagelist.get(i);
					Slot s_msg = buffer.getSlot(curr_seqn);
					if (s_msg != null)
						break;
					prev_seqn = curr_seqn;
				}
				/* Generate rejected message entry for missing messages */
				if (prev_seqn != -1) {
					RejectedMessage rm = new RejectedMessage(s, localmachineid, old_seqn, prev_seqn, false);
					s.addEntry(rm);
				}
				/* Generate rejected message entries for present messages */
				for (; i < rejectedmessagelist.size(); i++) {
					long curr_seqn = rejectedmessagelist.get(i);
					Slot s_msg = buffer.getSlot(curr_seqn);
					long machineid = s_msg.getMachineID();
					RejectedMessage rm = new RejectedMessage(s, machineid, curr_seqn, curr_seqn, true);
					s.addEntry(rm);
				}
			}
		}
	}

	private ThreeTuple<Boolean, Boolean, Long> doMandatoryResuce(Slot s, boolean resize) {
		long newestseqnum = buffer.getNewestSeqNum();
		long oldestseqnum = buffer.getOldestSeqNum();
		if (lastliveslotseqn < oldestseqnum)
			lastliveslotseqn = oldestseqnum;

		long seqn = lastliveslotseqn;
		boolean seenliveslot = false;
		long firstiffull = newestseqnum + 1 - numslots;	// smallest seq number in the buffer if it is full
		long threshold = firstiffull + FREE_SLOTS;	// we want the buffer to be clear of live entries up to this point


		// Mandatory Rescue
		for (; seqn < threshold; seqn++) {
			Slot prevslot = buffer.getSlot(seqn);
			// Push slot number forward
			if (! seenliveslot)
				lastliveslotseqn = seqn;

			if (! prevslot.isLive())
				continue;
			seenliveslot = true;
			Vector<Entry> liveentries = prevslot.getLiveEntries(resize);
			for (Entry liveentry : liveentries) {
				if (s.hasSpace(liveentry)) {
					s.addEntry(liveentry);
				} else if (seqn == firstiffull) { //if there's no space but the entry is about to fall off the queue
					if (!resize) {
						System.out.println("B"); //?
						return new ThreeTuple<Boolean, Boolean, Long>(true, seenliveslot, seqn);
					}
				}
			}
		}

		// Did not resize
		return new ThreeTuple<Boolean, Boolean, Long>(false, seenliveslot, seqn);
	}

	private boolean doArbitration(Slot s) {
		// Arbitrate
		Map<IoTString, KeyValue> speculativeTableTmp = new HashMap<IoTString, KeyValue>();
		List<Long> transSeqNums = new ArrayList<Long>(uncommittedTransactionsMap.keySet());

		// Sort from oldest to newest
		Collections.sort(transSeqNums);

		boolean didNeedArbitration = false;
		for (Long transNum : transSeqNums) {
			Transaction ut = uncommittedTransactionsMap.get(transNum);

			// Check if this machine arbitrates for this transaction
			if (ut.getArbitrator() != localmachineid ) {
				continue;
			}

			// we did have something to arbitrate on
			didNeedArbitration = true;

			Entry newEntry = null;

			if (ut.evaluateGuard(commitedTable, speculativeTableTmp)) {
				// Guard evaluated as true

				// update the local tmp current key set
				for (KeyValue kv : ut.getkeyValueUpdateSet()) {
					speculativeTableTmp.put(kv.getKey(), kv);
				}

				// create the commit
				newEntry = new Commit(s, ut.getSequenceNumber(), ut.getArbitrator(), ut.getkeyValueUpdateSet());
			} else {
				// Guard was false

				// create the abort
				newEntry = new Abort(s, ut.getSequenceNumber(), ut.getMachineID(), ut.getArbitrator());
			}

			if ((newEntry != null) && s.hasSpace(newEntry)) {
				s.addEntry(newEntry);
			} else {
				break;
			}
		}

		return didNeedArbitration;
	}

	private void  doOptionalRescue(Slot s, boolean seenliveslot, long seqn, boolean resize) {
		/* now go through live entries from least to greatest sequence number until
		 * either all live slots added, or the slot doesn't have enough room
		 * for SKIP_THRESHOLD consecutive entries*/
		int skipcount = 0;
		long newestseqnum = buffer.getNewestSeqNum();
		search:
		for (; seqn <= newestseqnum; seqn++) {
			Slot prevslot = buffer.getSlot(seqn);
			//Push slot number forward
			if (!seenliveslot)
				lastliveslotseqn = seqn;

			if (!prevslot.isLive())
				continue;
			seenliveslot = true;
			Vector<Entry> liveentries = prevslot.getLiveEntries(resize);
			for (Entry liveentry : liveentries) {
				if (s.hasSpace(liveentry))
					s.addEntry(liveentry);
				else {
					skipcount++;
					if (skipcount > SKIP_THRESHOLD)
						break search;
				}
			}
		}
	}

	private boolean doSendSlotsAndInsert(Slot s, boolean inserted, boolean resize, int newsize)  throws ServerException {
		int max = 0;
		if (resize)
			max = newsize;
		Slot[] array = cloud.putSlot(s, max);
		if (array == null) {
			array = new Slot[] {s};
			rejectedmessagelist.clear();
		}	else {
			if (array.length == 0)
				throw new Error("Server Error: Did not send any slots");
			rejectedmessagelist.add(s.getSequenceNumber());
			inserted = false;
		}

		validateandupdate(array, true);
		return inserted;
	}

	private void validateandupdate(Slot[] newslots, boolean acceptupdatestolocal) {
		/* The cloud communication layer has checked slot HMACs already
		         before decoding */
		if (newslots.length == 0) return;

		// Reset the table status declared sizes
		smallestTableStatusSeen = -1;
		largestTableStatusSeen = -1;

		long firstseqnum = newslots[0].getSequenceNumber();
		if (firstseqnum <= sequencenumber) {
			throw new Error("Server Error: Sent older slots!");
		}

		SlotIndexer indexer = new SlotIndexer(newslots, buffer);
		checkHMACChain(indexer, newslots);

		HashSet<Long> machineSet = new HashSet<Long>(lastmessagetable.keySet()); //

		// initExpectedSize(firstseqnum);
		for (Slot slot : newslots) {
			processSlot(indexer, slot, acceptupdatestolocal, machineSet);
			// updateExpectedSize();
		}

		/* If there is a gap, check to see if the server sent us everything. */
		if (firstseqnum != (sequencenumber + 1)) {

			// TODO: Check size
			checkNumSlots(newslots.length);
			if (!machineSet.isEmpty()) {
				throw new Error("Missing record for machines: " + machineSet);
			}
		}


		commitNewMaxSize();

		/* Commit new to slots. */
		for (Slot slot : newslots) {
			buffer.putSlot(slot);
			liveslotcount++;
		}
		sequencenumber = newslots[newslots.length - 1].getSequenceNumber();

		// Process all on key value pairs
		proccessAllNewCommits();

		// Go through all uncommitted transactions and kill the ones that are dead
		deleteDeadUncommittedTransactions();

		// Speculate on key value pairs
		createSpeculativeTable();
	}

	public void proccessAllNewCommits() {

		// Process only if there are commit
		if (newCommitMap.keySet().size() == 0) {
			return;
		}

		List<Long> commitSeqNums = new ArrayList<Long>(newCommitMap.keySet());

		// Sort from oldest to newest commit
		Collections.sort(commitSeqNums);

		// Go through each new commit one by one
		for (Long entrySeqNum : commitSeqNums) {
			Commit entry = newCommitMap.get(entrySeqNum);

			long lastCommitSeenSeqNum = 0;

			if (lastCommitSeenSeqNumMap.get(entry.getTransArbitrator()) != null) {
				lastCommitSeenSeqNum = lastCommitSeenSeqNumMap.get(entry.getTransArbitrator());
			}

			if (entry.getTransSequenceNumber() <= lastCommitSeenSeqNum) {

				Commit prevCommit = commitMap.put(entry.getTransSequenceNumber(), entry);

				if (prevCommit != null) {
					prevCommit.setDead();

					for (KeyValue kv : prevCommit.getkeyValueUpdateSet()) {
						committedMapByKey.put(kv.getKey(), entry);
					}
				}

				continue;
			}

			Set<Commit> commitsToEditSet = new HashSet<Commit>();

			for (KeyValue kv : entry.getkeyValueUpdateSet()) {
				commitsToEditSet.add(committedMapByKey.get(kv.getKey()));
			}

			commitsToEditSet.remove(null);

			for (Commit prevCommit : commitsToEditSet) {

				Set<KeyValue> deletedKV = prevCommit.updateLiveKeys(entry.getkeyValueUpdateSet());

				if (!prevCommit.isLive()) {
					commitMap.remove(prevCommit.getTransSequenceNumber());
				}
			}

			// // Remove any old commits
			// for (Iterator<Map.Entry<Long, Commit>> i = commitMap.entrySet().iterator(); i.hasNext();) {
			// 	Commit prevCommit = i.next().getValue();
			// 	prevCommit.updateLiveKeys(entry.getkeyValueUpdateSet());

			// 	if (!prevCommit.isLive()) {
			// 		i.remove();
			// 	}
			// }

			// Remove any old commits
			// for (Iterator<Map.Entry<Long, Commit>> i = commitMap.entrySet().iterator(); i.hasNext();) {
			// 	Commit prevCommit = i.next().getValue();

			// 	if (prevCommit.getTransArbitrator() != entry.getTransArbitrator()) {
			// 		continue;
			// 	}

			// 	prevCommit.updateLiveKeys(entry.getkeyValueUpdateSet());

			// 	if (!prevCommit.isLive()) {
			// 		i.remove();
			// 	}
			// }


			// Add the new commit
			commitMap.put(entry.getTransSequenceNumber(), entry);
			lastCommitSeenSeqNumMap.put(entry.getTransArbitrator(), entry.getTransSequenceNumber());

			// Update the committed table list
			for (KeyValue kv : entry.getkeyValueUpdateSet()) {
				IoTString key = kv.getKey();
				commitedTable.put(key, kv);

				committedMapByKey.put(key, entry);
			}
		}

		// Clear the new commits storage so we can use it later
		newCommitMap.clear();
	}

	private void deleteDeadUncommittedTransactions() {
		// Make dead the transactions
		for (Iterator<Map.Entry<Long, Transaction>> i = uncommittedTransactionsMap.entrySet().iterator(); i.hasNext();) {
			Transaction prevtrans = i.next().getValue();
			long transArb = prevtrans.getArbitrator();

			if ((lastCommitSeenSeqNumMap.get(transArb) != null) && (prevtrans.getSequenceNumber() <= lastCommitSeenSeqNumMap.get(transArb)) ||
			        (lastAbortSeenSeqNumMap.get(transArb) != null) && (prevtrans.getSequenceNumber() <= lastAbortSeenSeqNumMap.get(transArb))) {
				i.remove();
				prevtrans.setDead();
			}
		}
	}

	private void createSpeculativeTable() {

		if (uncommittedTransactionsMap.keySet().size() == 0) {
			// 	speculativeTable = commitedTable; // Ok that they are the same object
			return;
		}

		Map<IoTString, KeyValue> speculativeTableTmp = new HashMap<IoTString, KeyValue>();
		List<Long> utSeqNums = new ArrayList<Long>(uncommittedTransactionsMap.keySet());

		// Sort from oldest to newest commit
		Collections.sort(utSeqNums);

		if (utSeqNums.get(0) > (lastUncommittedTransaction)) {

			speculativeTable.clear();
			lastUncommittedTransaction = -1;

			for (Long key : utSeqNums) {
				Transaction trans = uncommittedTransactionsMap.get(key);

				lastUncommittedTransaction = key;

				if (trans.evaluateGuard(commitedTable, speculativeTableTmp)) {
					for (KeyValue kv : trans.getkeyValueUpdateSet()) {
						speculativeTableTmp.put(kv.getKey(), kv);
					}
				}

			}
		} else {
			for (Long key : utSeqNums) {

				if (key <= lastUncommittedTransaction) {
					continue;
				}

				lastUncommittedTransaction = key;

				Transaction trans = uncommittedTransactionsMap.get(key);

				if (trans.evaluateGuard(speculativeTable, speculativeTableTmp)) {
					for (KeyValue kv : trans.getkeyValueUpdateSet()) {
						speculativeTableTmp.put(kv.getKey(), kv);
					}
				}
			}
		}

		for (IoTString key : speculativeTableTmp.keySet()) {
			speculativeTable.put(key, speculativeTableTmp.get(key));
		}

		// speculativeTable = speculativeTableTmp;
	}

	private int expectedsize, currmaxsize;

	private void checkNumSlots(int numslots) {


		// We only have 1 size so we must have this many slots
		if (largestTableStatusSeen == smallestTableStatusSeen) {
			if (numslots != smallestTableStatusSeen) {
				throw new Error("Server Error: Server did not send all slots.  Expected: " + smallestTableStatusSeen + " Received:" + numslots);
			}
		} else {
			// We have more than 1
			if (numslots < smallestTableStatusSeen) {
				throw new Error("Server Error: Server did not send all slots.  Expected at least: " + smallestTableStatusSeen + " Received:" + numslots);
			}
		}

		// if (numslots != expectedsize) {
		// throw new Error("Server Error: Server did not send all slots.  Expected: " + expectedsize + " Received:" + numslots);
		// }
	}

	private void initExpectedSize(long firstsequencenumber) {
		long prevslots = firstsequencenumber;
		expectedsize = (prevslots < ((long) numslots)) ? (int) prevslots : numslots;
		currmaxsize = numslots;
	}

	private void updateExpectedSize() {
		expectedsize++;
		if (expectedsize > currmaxsize) {
			System.out.println("Maxing Out: " + expectedsize + "   " + currmaxsize);
			expectedsize = currmaxsize;
		}
	}

	private void updateCurrMaxSize(int newmaxsize) {
		currmaxsize = newmaxsize;
	}

	private void commitNewMaxSize() {

		if (largestTableStatusSeen == -1) {
			currmaxsize = numslots;
		} else {
			currmaxsize = largestTableStatusSeen;
		}

		if (numslots != currmaxsize) {
			buffer.resize(currmaxsize);
		}

		numslots = currmaxsize;
		setResizeThreshold();
	}

	private void processEntry(LastMessage entry, HashSet<Long> machineSet) {
		updateLastMessage(entry.getMachineID(), entry.getSequenceNumber(), entry, false, machineSet);
	}

	private void processEntry(RejectedMessage entry, SlotIndexer indexer) {
		long oldseqnum = entry.getOldSeqNum();
		long newseqnum = entry.getNewSeqNum();
		boolean isequal = entry.getEqual();
		long machineid = entry.getMachineID();
		for (long seqnum = oldseqnum; seqnum <= newseqnum; seqnum++) {
			Slot slot = indexer.getSlot(seqnum);
			if (slot != null) {
				long slotmachineid = slot.getMachineID();
				if (isequal != (slotmachineid == machineid)) {
					throw new Error("Server Error: Trying to insert rejected message for slot " + seqnum);
				}
			}
		}

		HashSet<Long> watchset = new HashSet<Long>();
		for (Map.Entry<Long, Pair<Long, Liveness> > lastmsg_entry : lastmessagetable.entrySet()) {
			long entry_mid = lastmsg_entry.getKey();
			/* We've seen it, don't need to continue to watch.  Our next
			 * message will implicitly acknowledge it. */
			if (entry_mid == localmachineid)
				continue;
			Pair<Long, Liveness> v = lastmsg_entry.getValue();
			long entry_seqn = v.getFirst();
			if (entry_seqn < newseqnum) {
				addWatchList(entry_mid, entry);
				watchset.add(entry_mid);
			}
		}
		if (watchset.isEmpty())
			entry.setDead();
		else
			entry.setWatchSet(watchset);
	}

	private void processEntry(NewKey entry) {
		arbitratorTable.put(entry.getKey(), entry.getMachineID());

		NewKey oldNewKey = newKeyTable.put(entry.getKey(), entry);

		if (oldNewKey != null) {
			oldNewKey.setDead();
		}
	}

	private void processEntry(Transaction entry) {

		long arb = entry.getArbitrator();
		Long comLast = lastCommitSeenSeqNumMap.get(arb);
		Long abLast = lastAbortSeenSeqNumMap.get(arb);

		Transaction prevTrans = null;

		if ((comLast != null) && (comLast >= entry.getSequenceNumber())) {
			prevTrans = uncommittedTransactionsMap.remove(entry.getSequenceNumber());
		} else if ((abLast != null) && (abLast >= entry.getSequenceNumber())) {
			prevTrans = uncommittedTransactionsMap.remove(entry.getSequenceNumber());
		} else {
			prevTrans = uncommittedTransactionsMap.put(entry.getSequenceNumber(), entry);
		}

		// Duplicate so delete old copy
		if (prevTrans != null) {
			prevTrans.setDead();
		}
	}

	private void processEntry(Abort entry) {

		if (lastmessagetable.get(entry.getMachineID()).getFirst() < entry.getTransSequenceNumber()) {
			// Abort has not been seen yet so we need to keep track of it

			Abort prevAbort = abortMap.put(entry.getTransSequenceNumber(), entry);
			if (prevAbort != null) {
				prevAbort.setDead(); // delete old version of the duplicate
			}
		} else {
			// The machine already saw this so it is dead
			entry.setDead();
		}

		if ((lastAbortSeenSeqNumMap.get(entry.getTransArbitrator()) != null) &&   (entry.getTransSequenceNumber() > lastAbortSeenSeqNumMap.get(entry.getTransArbitrator()))) {
			lastAbortSeenSeqNumMap.put(entry.getTransArbitrator(), entry.getTransSequenceNumber());
		}
	}

	private void processEntry(Commit entry, Slot s) {
		Commit prevCommit = newCommitMap.put(entry.getTransSequenceNumber(), entry);
		if (prevCommit != null) {
			prevCommit.setDead();
		}
	}

	private void processEntry(TableStatus entry) {
		int newnumslots = entry.getMaxSlots();
		// updateCurrMaxSize(newnumslots);
		if (lastTableStatus != null)
			lastTableStatus.setDead();
		lastTableStatus = entry;

		if ((smallestTableStatusSeen == -1) || (newnumslots < smallestTableStatusSeen)) {
			smallestTableStatusSeen = newnumslots;
		}

		if ((largestTableStatusSeen == -1) || (newnumslots > largestTableStatusSeen)) {
			largestTableStatusSeen = newnumslots;
		}

		// System.out.println("Table Stat: " + newnumslots + "   large: " + largestTableStatusSeen + "   small: " + smallestTableStatusSeen);
	}

	private void addWatchList(long machineid, RejectedMessage entry) {
		HashSet<RejectedMessage> entries = watchlist.get(machineid);
		if (entries == null)
			watchlist.put(machineid, entries = new HashSet<RejectedMessage>());
		entries.add(entry);
	}

	private void updateLastMessage(long machineid, long seqnum, Liveness liveness, boolean acceptupdatestolocal, HashSet<Long> machineSet) {
		machineSet.remove(machineid);

		HashSet<RejectedMessage> watchset = watchlist.get(machineid);
		if (watchset != null) {
			for (Iterator<RejectedMessage> rmit = watchset.iterator(); rmit.hasNext(); ) {
				RejectedMessage rm = rmit.next();
				if (rm.getNewSeqNum() <= seqnum) {
					/* Remove it from our watchlist */
					rmit.remove();
					/* Decrement machines that need to see this notification */
					rm.removeWatcher(machineid);
				}
			}
		}

		if (machineid == localmachineid) {
			/* Our own messages are immediately dead. */
			if (liveness instanceof LastMessage) {
				((LastMessage)liveness).setDead();
			} else if (liveness instanceof Slot) {
				((Slot)liveness).setDead();
			} else {
				throw new Error("Unrecognized type");
			}
		}

		// Set dead the abort
		for (Iterator<Map.Entry<Long, Abort>> i = abortMap.entrySet().iterator(); i.hasNext();) {
			Abort abort = i.next().getValue();

			if ((abort.getMachineID() == machineid) && (abort.getTransSequenceNumber() <= seqnum)) {
				abort.setDead();
				i.remove();
			}
		}

		Pair<Long, Liveness> lastmsgentry = lastmessagetable.put(machineid, new Pair<Long, Liveness>(seqnum, liveness));
		if (lastmsgentry == null)
			return;

		long lastmsgseqnum = lastmsgentry.getFirst();
		Liveness lastentry = lastmsgentry.getSecond();
		if (machineid != localmachineid) {
			if (lastentry instanceof LastMessage) {
				((LastMessage)lastentry).setDead();
			} else if (lastentry instanceof Slot) {
				((Slot)lastentry).setDead();
			} else {
				throw new Error("Unrecognized type");
			}
		}

		if (machineid == localmachineid) {
			if (lastmsgseqnum != seqnum && !acceptupdatestolocal)
				throw new Error("Server Error: Mismatch on local machine sequence number");
		} else {
			if (lastmsgseqnum > seqnum)
				throw new Error("Server Error: Rollback on remote machine sequence number");
		}
	}

	private void processSlot(SlotIndexer indexer, Slot slot, boolean acceptupdatestolocal, HashSet<Long> machineSet) {
		updateLastMessage(slot.getMachineID(), slot.getSequenceNumber(), slot, acceptupdatestolocal, machineSet);
		for (Entry entry : slot.getEntries()) {
			switch (entry.getType()) {

			case Entry.TypeNewKey:
				processEntry((NewKey)entry);
				break;

			case Entry.TypeCommit:
				processEntry((Commit)entry, slot);
				break;

			case Entry.TypeAbort:
				processEntry((Abort)entry);
				break;

			case Entry.TypeTransaction:
				processEntry((Transaction)entry);
				break;

			case Entry.TypeLastMessage:
				processEntry((LastMessage)entry, machineSet);
				break;

			case Entry.TypeRejectedMessage:
				processEntry((RejectedMessage)entry, indexer);
				break;

			case Entry.TypeTableStatus:
				processEntry((TableStatus)entry);
				break;

			default:
				throw new Error("Unrecognized type: " + entry.getType());
			}
		}
	}

	private void checkHMACChain(SlotIndexer indexer, Slot[] newslots) {
		for (int i = 0; i < newslots.length; i++) {
			Slot currslot = newslots[i];
			Slot prevslot = indexer.getSlot(currslot.getSequenceNumber() - 1);
			if (prevslot != null &&
			        !Arrays.equals(prevslot.getHMAC(), currslot.getPrevHMAC()))
				throw new Error("Server Error: Invalid HMAC Chain" + currslot + " " + prevslot);
		}
	}
}
